from textwrap import dedent
from typing import Callable, List, Mapping, Sequence

from . import types


class UndefinedProfile(ValueError):
    """The given profile ('{name}') is not registered with ``ini2toml``.
    Are you sure you have the right plugins installed and loaded?
    """

    def __init__(self, name: str, available: Sequence[str]):
        msg = self.__class__.__doc__ or ""
        super().__init__(msg.format(name=name) + f"Available: {', '.join(available)})")

    @classmethod
    def check(cls, name: str, available: List[str]):
        """:meta private:"""
        if name not in available:
            raise cls(name, available)


class AlreadyRegisteredAugmentation(ValueError):
    """The profile augmentation '{name}' is already registered for '{existing}'.

    Some installed plugins seem to be in conflict with each other,
    please check '{new}' and '{existing}'.
    If you are the developer behind one of them, please use a different name.
    """

    def __init__(self, name: str, new: Callable, existing: Callable):
        existing_id = f"{existing.__module__}.{existing.__qualname__}"
        new_id = f"{new.__module__}.{new.__qualname__}"
        msg = dedent(self.__class__.__doc__ or "")
        super().__init__(msg.format(name=name, new=new_id, existing=existing_id))

    @classmethod
    def check(
        cls, name: str, fn: Callable, registry: Mapping[str, types.ProfileAugmentation]
    ):
        """:meta private:"""
        if name in registry:
            raise cls(name, fn, registry[name].fn)


class InvalidAugmentationName(ValueError):
    """Profile augmentations should be valid python identifiers"""

    def __init__(self, name: str):
        msg = self.__class__.__doc__ or ""
        super().__init__(f"{msg} ({name!r} given)")

    @classmethod
    def check(cls, name: str):
        if not name.isidentifier():
            raise cls(name)


class InvalidTOMLKey(ValueError):
    """{key!r} is not a valid key in the intermediate TOML representation"""

    def __init__(self, key):
        msg = self.__doc__.format(key=key)
        super().__init__(msg)


class InvalidCfgBlock(ValueError):  # pragma: no cover -- not supposed to happen
    """Something is wrong with the provided ``.ini/.cfg`` AST"""

    def __init__(self, block):
        super().__init__(f"{block.__class__}: {block}", {"block_object": block})



from types import MappingProxyType
from typing import Mapping

from configupdater import Comment, ConfigUpdater, Option, Section, Space

from ..errors import InvalidCfgBlock
from ..transformations import remove_prefixes
from ..types import CommentKey, IntermediateRepr, WhitespaceKey

EMPTY: Mapping = MappingProxyType({})
COMMENT_PREFIXES = ("#", ";")


def parse(text: str, opts: Mapping = EMPTY) -> IntermediateRepr:
    cfg = ConfigUpdater(**opts).read_string(text)
    irepr = IntermediateRepr()

    for block in cfg.iter_blocks():
        if isinstance(block, Section):
            translate_section(irepr, block, opts)
        elif isinstance(block, Comment):
            translate_comment(irepr, block, opts)
        elif isinstance(block, Space):
            translate_space(irepr, block, opts)
        else:  # pragma: no cover -- not supposed to happen
            raise InvalidCfgBlock(block)

    return irepr


def translate_section(irepr: IntermediateRepr, item: Section, opts: Mapping):
    out = IntermediateRepr()
    # Inline comment
    cmt = getattr(item, "_raw_comment", "")  # TODO: private attr
    cmt = remove_prefixes(cmt, opts.get("comment_prefixes", COMMENT_PREFIXES))
    if cmt:
        out.inline_comment = cmt
    # Children
    for block in item.iter_blocks():
        if isinstance(block, Option):
            translate_option(out, block, opts)
        elif isinstance(block, Comment):
            translate_comment(out, block, opts)
        elif isinstance(block, Space):
            translate_space(out, block, opts)
        else:  # pragma: no cover -- not supposed to happen
            raise InvalidCfgBlock(block)
    irepr.append(item.name, out)


def translate_option(irepr: IntermediateRepr, item: Option, _opts: Mapping):
    irepr.append(item.key, item.value)


def translate_comment(irepr: IntermediateRepr, item: Comment, opts: Mapping):
    prefixes = opts.get("comment_prefixes", COMMENT_PREFIXES)
    for line in str(item).splitlines():
        irepr.append(CommentKey(), remove_prefixes(line, prefixes))


def translate_space(irepr: IntermediateRepr, item: Space, _opts: Mapping):
    for line in str(item).splitlines(keepends=True):
        irepr.append(WhitespaceKey(), line)



"""Profile-independent tasks implemented via *profile augmentation*."""

import re
from functools import wraps
from typing import Callable

from ..types import Profile, Translator

DUPLICATED_NEWLINES = re.compile(r"\n+", re.M)
TABLE_START = re.compile(r"^\[(.*)\]", re.M)
EMPTY_TABLES = re.compile(r"^\[(.*)\]\n+\[(\1\.(?:.*))\]", re.M)
MISSING_TERMINATING_LINE = re.compile(r"(?<!\n)\Z", re.M)
# ^ POSIX tools will not play nicely with text files without a terminating new line
# https://unix.stackexchange.com/questions/18743/whats-the-point-in-adding-a-new-line-to-the-end-of-a-file


def activate(translator: Translator):
    tasks = [
        normalise_newlines,
        remove_empty_table_headers,
        ensure_terminating_newlines,
    ]
    for task in tasks:
        translator.augment_profiles(post_process(task), active_by_default=True)


def post_process(fn: Callable[[str], str]):
    @wraps(fn)
    def _augmentation(profile: Profile):
        profile.post_processors.append(fn)

    return _augmentation


def normalise_newlines(text: str) -> str:
    """Make sure every table is preceded by an empty newline, but remove them elsewhere
    in the output TOML document.
    Also ensure a terminating newline is present for best POSIX tool compatibility.
    """
    text = DUPLICATED_NEWLINES.sub(r"\n", text)
    text = TABLE_START.sub(r"\n[\1]", text)
    return MISSING_TERMINATING_LINE.sub("\n", text)


def remove_empty_table_headers(text: str) -> str:
    """Remove empty TOML table headers"""
    prev_text = ""
    while text != prev_text:
        prev_text = text
        text = EMPTY_TABLES.sub(r"[\2]", text).strip()
    return text


def ensure_terminating_newlines(text: str) -> str:
    return text.strip() + "\n"



"""This module serves as a compatibility layer between API-compatible
TOML parsers/serialisers.
It makes it easy to swap between implementations for testing (by means of search and
replace).
"""

try:
    from tomli_w import dumps
except ImportError:  # pragma: no cover
    # Let's try another API-compatible popular library as a last hope
    from toml import dumps  # type: ignore[import-untyped,no-redef]

from ..types import IntermediateRepr
from . import plain_builtins

__all__ = [
    "convert",
]


def convert(irepr: IntermediateRepr) -> str:
    text = dumps(plain_builtins.convert(irepr))
    return text.strip() + "\n"  # ensure terminating newline (POSIX requirement)



import logging
from inspect import cleandoc
from typing import List, TypeVar

from ..types import IntermediateRepr, Translator

R = TypeVar("R", bound=IntermediateRepr)

_logger = logging.getLogger(__package__)

_FLAKE8_SECTIONS = ["flake8", "flake8-rst", "flake8:local-plugins"]

_KNOWN_INCOMPATIBILITIES = {
    "setup.cfg": [*_FLAKE8_SECTIONS, "devpi:upload"],
    ".flake8": _FLAKE8_SECTIONS,
}


def activate(translator: Translator):
    for name, sections in _KNOWN_INCOMPATIBILITIES.items():
        fn = ReportIncompatibleSections(name, sections)
        translator[name].intermediate_processors.insert(0, fn)


class ReportIncompatibleSections:
    """Remove well-know incompatible sections."""

    def __init__(self, profile: str, sections: List[str]):
        self._profile = profile
        self._sections = sections

    def __call__(self, cfg: R) -> R:
        invalid = [section for section in self._sections if section in cfg]
        if invalid:
            sections = ", ".join(repr(x) for x in invalid)
            _logger.warning(_warning_text(self._profile, sections))
        return cfg


def _warning_text(profile: str, sections: str) -> str:
    msg = f"""
    Sections {sections} ({profile!r}) may be problematic.

    It might be the case TOML files are not supported by the relevant tools,
    or that 'ini2toml' just lacks a plugin for this kind of configuration.

    Please review the generated output and remove these sections if necessary.
    """
    return cleandoc(msg) + "\n"



"""Public API available for general usage.

In addition to the classes and functions "exported" by this module, the following are
also part of the public API:

- The public members of the :mod:`~ini2toml.types` module.
- The public members of the :mod:`~ini2toml.errors` module.
- The ``activate`` function in each submodule of the :obj:`~ini2toml.plugins` package

Please notice there might be classes of similar names exported by both ``api`` and
``types``. When this happens, the classes in ``types`` are not concrete implementations,
but instead act as :class:`protocols <typing.Protocol>` (i.e. abstract descriptions
for checking `structural polymorphism`_ during static analysis).
These should be preferred when writing type hints and signatures.

Plugin authors can also use functions exported by :mod:`~ini2toml.transformations`.

.. _structural polymorphism: https://www.python.org/dev/peps/pep-0544/
"""

from .base_translator import BaseTranslator
from .translator import FullTranslator, LiteTranslator, Translator

__all__ = [
    "BaseTranslator",
    "FullTranslator",
    "LiteTranslator",
    "Translator",
]



# https://pycqa.github.io/isort/docs/configuration/config_files
from collections.abc import Mapping, MutableMapping
from functools import partial, update_wrapper
from typing import Set, TypeVar

from ..transformations import coerce_scalar, kebab_case, split_list
from ..types import Transformation as T
from ..types import Translator

M = TypeVar("M", bound=MutableMapping)


def activate(translator: Translator):
    plugin = ISort()
    profile = translator[".isort.cfg"]
    fn = partial(plugin.process_values, section_name="settings")
    update_wrapper(fn, plugin.process_values)
    profile.intermediate_processors.append(fn)
    profile.help_text = plugin.__doc__ or ""

    for file in ("setup.cfg", "tox.ini"):
        translator[file].intermediate_processors.append(plugin.process_values)


class ISort:
    """Convert settings to 'pyproject.toml' equivalent"""

    FIELDS = {
        "force_to_top",
        "treat_comments_as_code",
        "no_lines_before",
        "forced_separate",
        "sections",
        "length_sort_sections",
        "sources",
        "constants",
        "classes",
        "variables",
        "namespace_packages",
        "add_imports",
        "remove_imports",
    }

    FIELD_ENDS = ["skip", "glob", "paths", "exclusions", "plugins"]
    FIELD_STARTS = ["known", "extra"]

    # dicts? ["import_headings", "git_ignore", "know_other"]

    def find_list_options(self, section: Mapping) -> Set[str]:
        dynamic_fields = (
            field
            for field in section
            if isinstance(field, str)
            and (
                any(field.startswith(s) for s in self.FIELD_STARTS)
                or any(field.endswith(s) for s in self.FIELD_ENDS)
            )
        )
        fields = {*self.FIELDS, *dynamic_fields}
        return {*fields, *map(kebab_case, fields)}

    def process_values(self, doc: M, section_name="isort") -> M:
        candidates = [
            doc.get(section_name),
            doc.get(("tool", section_name)),
            doc.get("tool", {}).get(section_name),
        ]
        for section in candidates:
            if section:
                self.process_section(section)
        return doc

    def process_section(self, section: M):
        list_options = self.find_list_options(section)
        for field in section:
            fn: T = split_list if field in list_options else coerce_scalar
            section[field] = fn(section[field])



import inspect
from copy import deepcopy
from typing import Optional, Sequence, TypeVar

from .types import IntermediateProcessor, ProfileAugmentationFn, TextProcessor

P = TypeVar("P", bound="Profile")


def replace(self: P, **changes) -> P:
    """Works similarly to :func:`dataclasses.replace`"""
    sig = inspect.signature(self.__class__)
    kwargs = {x: getattr(self, x) for x in sig.parameters}
    kwargs.update(changes)
    return self.__class__(**kwargs)


class Profile:
    """Profile object that follows the public API defined in
    :class:`ini2toml.types.Profile`.
    """

    def __init__(
        self,
        name: str,
        help_text: str = "",
        pre_processors: Sequence[TextProcessor] = (),
        intermediate_processors: Sequence[IntermediateProcessor] = (),
        post_processors: Sequence[TextProcessor] = (),
        ini_parser_opts: Optional[dict] = None,
    ):
        self.name = name
        self.help_text = help_text
        self.pre_processors = list(pre_processors)
        self.intermediate_processors = list(intermediate_processors)
        self.post_processors = list(post_processors)
        self.ini_parser_opts = ini_parser_opts

    replace = replace

    def _copy(self: P) -> P:
        return self.__class__(
            name=self.name,
            help_text=self.help_text,
            pre_processors=self.pre_processors[:],
            intermediate_processors=self.intermediate_processors[:],
            post_processors=self.post_processors[:],
            ini_parser_opts=deepcopy(self.ini_parser_opts),
        )


class ProfileAugmentation:
    def __init__(
        self,
        fn: ProfileAugmentationFn,
        active_by_default: bool = False,
        name: str = "",
        help_text: str = "",
    ):
        self.fn = fn
        self.active_by_default = active_by_default
        self.name = name or getattr(fn, "__name__", "")
        self.help_text = help_text or getattr(fn, "__doc__", "")

    def is_active(self, explicitly_active: Optional[bool] = None) -> bool:
        """``explicitly_active`` is a tree-state variable: ``True`` if the user
        explicitly asked for the augmentation, ``False`` if the user explicitly denied
        the augmentation, or ``None`` otherwise.
        """
        activation = explicitly_active
        return activation is True or (activation is None and self.active_by_default)



import logging
import re
import warnings
from functools import partial, reduce
from itertools import chain, zip_longest
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

try:
    from packaging.requirements import Requirement
except ImportError:  # pragma: no cover
    from setuptools.extern.packaging.requirements import Requirement  # type: ignore

from distutils import command as distutils_commands

from ..transformations import (
    apply,
    coerce_bool,
    deprecated,
    kebab_case,
    noop,
    split_comment,
    split_kv_pairs,
    split_list,
)
from ..types import Commented, CommentedKV, CommentedList, CommentKey, HiddenKey
from ..types import IntermediateRepr as IR
from ..types import Transformation, Translator, WhitespaceKey
from .best_effort import BestEffort

R = TypeVar("R", bound=IR)

RenameRules = Dict[Tuple[str, ...], Union[Tuple[Union[str, int], ...], None]]
ProcessingRules = Dict[Tuple[str, ...], Transformation]


_logger = logging.getLogger(__name__)

chain_iter = chain.from_iterable

ENV_MARKER = re.compile(r";\s*(python|platform|implementation|os|sys)[_.]", re.M)
"""Simplified regex for :pep:`508` markers that can be used in ``setup.cfg``"""

# Functions that split values from comments and parse those values
split_list_comma = partial(split_list, sep=",", subsplit_dangling=False)
split_list_semi = partial(split_list, sep=";", subsplit_dangling=False)
split_hash_comment = partial(split_comment, comment_prefixes="#")  # avoid splitting `;`
split_bool = partial(split_comment, coerce_fn=coerce_bool)
split_kv_of_lists = partial(split_kv_pairs, coerce_fn=split_list_comma)
split_keywords = partial(split_list_comma, coerce_fn=lambda x: x.rstrip(","))
# URLs can contain the # symbol
split_kv_urls = partial(split_kv_pairs, comment_prefixes=(" #",))
split_url = partial(split_comment, comment_prefixes=(" #",))

SECTION_SPLITTER = re.compile(r"\.|:")
SETUPTOOLS_SECTIONS = ("metadata", "options")
SKIP_CHILD_NORMALISATION = (
    "options.entry_points",
    "options.package_data",
    "options.exclude_package_data",
    "options.extras_require",
    "options.data_files",
)
COMMAND_SECTIONS = (
    "global",
    "alias",
    "install",
    "develop",
    "sdist",
    "bdist",
    "bdist_wheel",
    *getattr(distutils_commands, "__all__", []),
)
SKIP_COMMAND_SECTIONS = {
    "isort",
}


def activate(translator: Translator):
    plugin = SetuptoolsPEP621()
    profile = translator["setup.cfg"]
    profile.intermediate_processors += [plugin.normalise_keys, plugin.pep621_transform]
    profile.help_text = plugin.__doc__ or ""


class SetuptoolsPEP621:
    """Convert settings to 'pyproject.toml' based on :pep:`621`"""

    BUILD_REQUIRES = ("setuptools>=61.2",)

    def __init__(self):
        self._be = BestEffort(key_sep="=")

    @classmethod
    def template(
        cls,
        ir_cls: Type[R] = IR,  # type: ignore
        build_requires: Sequence[str] = (),
    ) -> R:
        build_system = {
            "requires": [*(build_requires or cls.BUILD_REQUIRES)],
            # ^ NOTE: the code ahead assumes no version
            "build-backend": "setuptools.build_meta",
        }
        tpl = {
            "build-system": ir_cls(build_system),  # type: ignore
            "metadata": ir_cls(),  # NOTE: will be renamed later
            "tool": ir_cls(),
        }
        return ir_cls(tpl)  # type: ignore

    def setupcfg_aliases(self):
        """``setup.cfg`` aliases as defined in:
        https://setuptools.pypa.io/en/stable/userguide/declarative_config.html
        """
        return {
            "classifier": "classifiers",
            "summary": "description",
            "platform": "platforms",
            "license-file": "license-files",
            "home-page": "url",
        }

    def processing_rules(self) -> ProcessingRules:
        """Value type processing, as defined in:
        https://setuptools.pypa.io/en/stable/userguide/declarative_config.html
        """
        # If not present below will be transformed via split_comment by default
        return {
            ("metadata", "version"): directive("file", "attr"),
            ("metadata", "classifiers"): directive("file", orelse=split_list_comma),
            ("metadata", "keywords"): split_keywords,
            ("metadata", "description"): directive("file"),
            # ---
            ("metadata", "long-description"): directive("file", orelse=noop),
            ("metadata", "long-description-content-type"): split_hash_comment,
            # => NOTICE: further processed via
            #            `merge_and_rename_long_description_and_content_type`
            # ---
            ("metadata", "license-files"): split_list_comma,
            # => NOTICE: not standard for now, needs PEP 639
            #            further processed via `remove_metadata_not_in_pep621`
            # ---
            ("metadata", "url"): split_url,
            ("metadata", "download-url"): split_url,
            ("metadata", "project-urls"): split_kv_urls,
            # => NOTICE: further processed via `merge_and_rename_urls`
            # ---- Not covered by PEP 621 ----
            ("metadata", "platforms"): split_list_comma,
            # ---
            ("metadata", "provides"): split_list_comma,
            ("metadata", "requires"): deprecated("requires", split_list_comma),
            ("metadata", "obsoletes"): split_list_comma,
            # => NOTICE: not supported by pip
            # ---- Options ----
            ("options", "zip-safe"): split_bool,
            ("options", "setup-requires"): split_deps,
            ("options", "install-requires"): directive("file", orelse=split_deps),
            ("options", "tests-require"): split_deps,
            ("options", "scripts"): split_list_comma,
            ("options", "eager-resources"): split_list_comma,
            ("options", "dependency-links"): deprecated(
                "dependency-links", split_list_comma
            ),  # noqa
            ("options", "entry-points"): directive(
                "file", orelse=value_error("option.entry-points")
            ),
            ("options", "include-package-data"): split_bool,
            ("options", "package-dir"): split_kv_pairs,
            ("options", "namespace-packages"): split_list_comma,
            ("options", "py-modules"): split_list_comma,
            ("options", "cmdclass"): split_kv_pairs,
            ("options", "data-files"): deprecated("data-files", split_kv_of_lists),
            ("options", "packages"): directive(
                "find", "find_namespace", orelse=split_list_comma
            ),
            ("options.packages.find", "include"): split_list_comma,
            ("options.packages.find", "exclude"): split_list_comma,
            ("options.packages.find", "exclude"): split_list_comma,
        }
        # See also dependent_processing_rules

    def dependent_processing_rules(self, doc: IR) -> ProcessingRules:
        """Dynamically create processing rules, such as :func:`processing_rules` based
        on the existing document.
        """
        groups: Mapping[str, Transformation] = {
            "options.extras-require": directive("file", orelse=split_deps),
            "options.package-data": split_list_comma,
            "options.exclude-package-data": split_list_comma,
            "options.data-files": split_list_comma,
            "options.entry-points": split_kv_pairs,
        }
        return {
            (g, k): fn
            for g, fn in groups.items()
            for k in doc.get(g, ())
            if isinstance(k, str)
        }

    def apply_value_processing(self, doc: R) -> R:
        """Process ``setup.cfg`` values according to :meth:`processing_rules` and
        :meth:`dependent_processing_rules`.

        This function assumes all field names were normalised by :meth:`normalise_keys`.
        """
        default = {
            (name, option): split_comment
            for name, section in doc.items()
            if name in ("metadata", "options")
            for option in section
            if isinstance(option, (str, tuple)) and not isinstance(option, HiddenKey)
        }
        transformations: dict = {
            **default,
            **self.processing_rules(),
            **self.dependent_processing_rules(doc),
        }
        for (section, option), fn in transformations.items():
            value = doc.get(section, {}).get(option, None)
            if value is not None:
                doc[section][option] = fn(value)
        return doc

    def merge_and_rename_urls(self, doc: R) -> R:
        """The following renames can be applied when comparing setuptools metadata and
        :pep:`621`::

            url => urls.homepage
            download-url => urls.download
            project-urls.* => urls.*
        """
        metadata: IR = doc["metadata"]
        new_urls = [
            (dest, metadata.pop(orig))
            for orig, dest in [("url", "Homepage"), ("download-url", "Download")]
            if orig in metadata
        ]
        urls = metadata.get("project-urls", CommentedKV())
        for k, v in reversed(new_urls):
            urls.insert_line(0, [(k, v.value)], v.comment)

        if urls.as_dict():
            keys = ("project-urls", "url", "download-url")
            metadata.replace_first_remove_others(keys, "urls", urls)
        return doc

    def merge_authors_maintainers_and_emails(self, doc: R) -> R:
        """When transforming setuptools metadata and :pep:`621`, we have to merge
        ``author/maintainer`` and ``author-email/maintainer-email`` into a single
        dict-like object with 2 keys.
        Some projects also provide multiple, comma separated, values for each field.
        In that case we assume that the value for the i-th author/maintainer should be
        paired with to the i-th author-email/maintainer-email value.
        """
        metadata: IR = doc["metadata"]

        def _split_values(field: str) -> Tuple[Iterator[str], Optional[str]]:
            commented: Commented[str] = metadata.get(field, Commented())
            values = commented.value_or("").strip().split(",")
            return (v.strip() for v in values), commented.comment

        for key in ("author", "maintainer"):
            fields = (key, f"{key}-email")
            values, comments = zip(*(_split_values(f) for f in fields))
            combined = (
                {k: v for k, v in zip(("name", "email"), person_data) if v}
                # ^-- Remove empty fields
                for person_data in zip_longest(*values, fillvalue="")
            )
            people = [IR(c) for c in combined if c]  # type: ignore[arg-type]

            if people:
                # author/maintainer => author**S**/maintainer**S**
                i = metadata.replace_first_remove_others(fields, f"{key}s", people)
                for j, cmt in enumerate(c for c in comments if c):
                    metadata.insert(j + i + 1, CommentKey(), cmt)
        return doc

    def merge_and_rename_long_description_and_content_type(self, doc: R) -> R:
        """:pep:`621` offers a single field (``readme``) to cover things present in two
        fields in ``setup.cfg``::

            long_description.file => readme.file
            long_description => readme.text
            long-description-content-type => readme.content-type

        We also have to be aware that :pep:`621` accepts a single file, so the option of
        combining multiple files as presented in ``setup.cfg`` have to be handled via
        ``dynamic``.
        """
        metadata: IR = doc["metadata"]
        long_desc: Union[Directive, str, None] = metadata.get("long-description")
        if not long_desc:
            metadata.pop("long-description", None)
            metadata.pop("long-description-content-type", None)
            return doc

        readme: Dict[str, Any] = {}
        dynamic = False
        if isinstance(long_desc, Directive):
            # In PEP 621 "readme" should be a single file
            files: CommentedList[str] = long_desc["file"]
            files_list = files.as_list()
            if len(files_list) == 1:
                readme = {"file": Commented(files_list[0], files[0].comment)}
            else:
                readme = dict(long_desc)
                dynamic = True
        else:
            readme = {"text": long_desc.strip()}

        content_type = metadata.pop("long-description-content-type", None)
        if content_type:
            readme["content-type"] = content_type

        if dynamic:
            metadata.setdefault("dynamic", []).append("readme")
            doc.setdefault("options.dynamic", IR()).append("readme", readme)
            metadata.pop("long-description")
            return doc

        if len(list(readme.keys())) == 1 and "file" in readme:
            metadata["long-description"] = readme["file"]
        else:
            metadata["long-description"] = IR(readme)  # type: ignore[arg-type]
        metadata.rename("long-description", "readme")
        return doc

    def handle_license(self, doc: R) -> R:
        """In :pep:`621` we have a single field for license, which is not compatible
        with setuptools ``license-files``.
        This field is meant to fill the ``License`` core metadata as a plain license
        text (not a path to a file). Even when the ``project.license.file`` table field
        is given, the idea is that the file should be expanded into text.

        This will likely change once :pep:`639` is accepted.
        Meanwhile we have to translate ``license-files`` into a setuptools specific
        configuration.
        """
        metadata: IR = doc["metadata"]
        if "license" in metadata:
            metadata.rename("license", ("license", "text"))
        return doc

    def move_and_split_entrypoints(self, doc: R) -> R:
        """In ``setup.cfg`` there is no special treatment for entry-points that will be
        transformed in console/GUI scripts. On the other hand :pep:`621` defines
        separated fields::

            entry-points.console-scripts => scripts
            entry-points.gui-scripts => gui-scripts
            entry-points.* => "entry-points".*
        """
        entrypoints: IR = doc.get("options.entry-points", IR())
        if not entrypoints:
            doc.pop("options.entry-points", None)
            return doc
        doc.rename("options.entry-points", "project:entry-points")
        # ^ use `:` to guarantee it is split later
        script_keys = ["console-scripts", "gui-scripts"]
        script_keys += [k.replace("-", "_") for k in script_keys]
        keys = (k for k in script_keys if k in entrypoints)
        for key in keys:
            scripts: CommentedKV = entrypoints.pop(key)
            new_key = key.replace("_", "-").replace("console-", "")
            doc.append(f"project:{new_key}", scripts.to_ir())
        if not entrypoints or all(isinstance(k, WhitespaceKey) for k in entrypoints):
            doc.pop("project:entry-points")
        return doc

    def move_options_missing_in_pep621(self, doc: R) -> R:
        """:pep:`621` specifies as project metadata values that are covered
        in ``setup.cfg "options"`` section.
        """
        # First we handle simple options
        naming = {
            "python-requires": "requires-python",
            "install-requires": "dependencies",
            "entry-points": "entry-points",
        }
        metadata = doc.setdefault("metadata", IR())
        options = doc.setdefault("options", IR())
        metadata.update({v: options.pop(k) for k, v in naming.items() if k in options})

        # Then we handle entire sections:
        naming = {"extras-require": "optional-dependencies"}
        for src, target in naming.items():
            doc.rename(f"options.{src}", f"project:{target}", ignore_missing=True)

        return doc

    def remove_metadata_not_in_pep621(self, doc: R) -> R:
        """:pep:`621` does not cover all project metadata in ``setup.cfg "metadata"``
        section. That is left as "tool" specific configuration.
        """
        # TODO: PEP 621 does not specify an equivalent for 'License-file' metadata,
        #       but once PEP 639 is approved this will change
        metadata = doc.get("metadata", IR())
        non_standard = ("platforms", "provides", "obsoletes", "license-files")
        specific = [k for k in non_standard if k in metadata]
        if specific:
            options = doc.setdefault("options", IR())
            options.update({k: metadata.pop(k) for k in specific})
        return doc

    def rename_script_files(self, doc: R) -> R:
        """``setuptools`` define a ``options.scripts`` parameters that refer to
        script files, not created via entry-points.
        To avoid confusion with :pep:`621` scripts (generated via entry-points)
        let's rename this field to ``script-files``
        """
        doc["options"].rename("scripts", "script-files", ignore_missing=True)
        return doc

    def handle_packages_find(self, doc: R) -> R:
        """``setup.cfg`` uses a option + a section to define ``options.packages.find``
        and its ``find_namespace`` variant. This does not work very well with the
        convention used for the TOML encoding, since the option and the section would
        end up with the same name (and overwriting each other). Therefore we need to
        "merge" them.
        """
        options = doc["options"]
        # Abort when not using find or find_namespace
        packages = options.get("packages")

        if not isinstance(packages, Directive):
            if "options.packages.find" in doc:
                _ConfusingPackagesConfig.emit()
                doc.pop("options.packages.find", None)
            return doc

        prefix = packages.kind.replace("_", "-")
        # Enhancement #1: Unify find and find_namespaces, using `namespaces` as a flag
        options["packages"] = Directive("find", {"namespaces": "namespace" in prefix})
        if "options.packages.find" in doc:
            packages = options.pop("packages")
            doc["options.packages.find"].update(packages["find"])
            # Enhancement #2: ``where`` accepts multiple values (array)
            where = doc["options.packages.find"].get("where", None)
            if where:
                doc["options.packages.find"]["where"] = _ensure_where_list(where)
        return doc

    def handle_dynamic(self, doc: R) -> R:
        """All the configuration fields in :pep:`621` that are dynamically discovered at
        build time have to be explicitly list in ``dynamic``.
        This function moves directive usages (e.g. ``file:`` and ``attr:``) to a
        tool-specific subtable (``tool.setuptools.dynamic``), and add the corresponding
        field to ``dynamic``.
        Since ``version`` is a mandatory core metadata, it will be added to ``dynamic``
        when not explicitly set (in that case plugins such as ``setuptools_scm`` are
        expected to provide a value at runtime).
        """
        potential = ["version", "classifiers", "description", "dependencies"]
        # directives = {k[-1]: v for k, v in self.setupcfg_directives().items()}
        metadata, options = doc["metadata"], doc["options"]

        field_falues = ((f, metadata.get(f)) for f in potential)
        fields = [f for f, v in field_falues if isinstance(v, Directive)]

        dynamic = {f: metadata.pop(f, None) for f in fields}
        if "version" not in metadata and "version" not in dynamic:
            msg = (
                "No `version` was found in `[metadata]`, `ini2toml` will assume it is "
                "defined by tools like `setuptools-scm` or in `setup.py`. "
                "Automatically adding it to `dynamic` (in accordance with PEP 621)"
            )
            _logger.debug(msg)
            fields.insert(0, "version")

        extras: List[str] = []
        ep = metadata.pop("entry-points", options.pop("entry-points", None))
        if isinstance(ep, Directive):
            fields.append("entry-points")
            dynamic["entry-points"] = ep
            extras = ["scripts", "gui-scripts"]
        if not fields:
            return doc

        dyn_deps = [
            isinstance(v, Directive)
            for v in doc.get("project:optional-dependencies", {}).values()
        ]
        if any(dyn_deps):
            dynamic["optional-dependencies"] = doc.pop("project:optional-dependencies")
            fields.append("optional-dependencies")
            if not all(dyn_deps):
                _logger.warning(
                    "setuptools may require all optional-dependencies to be dynamic "
                    "when using `pyproject.toml`."
                )

        metadata.setdefault("dynamic", []).extend(fields + extras)
        if dynamic:
            doc.setdefault("options.dynamic", IR()).update(dynamic)
            # ^ later `options.dynamic` is converted to `tool.setuptools.dynamic`
        return doc

    def fix_extras_require(self, doc: R) -> R:
        """`extras-require` can have markers embedded in the extra group
        they need to be removed and added to the dependencies themselves
        """
        if "project:optional-dependencies" not in doc:
            return doc

        extras = doc["project:optional-dependencies"]
        keys = list(extras.keys())  # Eager, so we can modify extras
        for key in keys:
            if not isinstance(key, str):
                continue
            extra_name, _, marker = key.partition(":")
            extra_name, marker = extra_name.strip(), marker.strip()
            if not marker:
                continue
            values = extras[key]
            extras.rename(key, extra_name)
            extras[extra_name] = [_add_marker(r, marker) for r in values.as_list()]

        return doc

    def move_setup_requires(self, doc: R) -> R:
        """Move ``setup_requires`` to the equivalent field in :pep:`518`, and add
        mandatory build dependencies if they are missing and
        """
        options = doc["options"]
        build_system = doc["build-system"]
        if "setup-requires" in options:
            msg = "The field 'setup_requires' is deprecated. "
            msg += "Converting to `build-system.requires` as specified by PEP 518."
            warnings.warn(msg, DeprecationWarning)  # noqa: B028
            requirements: CommentedList[str] = options.pop("setup-requires")
            # Deduplicate
            existing = {Requirement(r).name: r for r in requirements.as_list()}
            mandatory = {
                Requirement(r).name: r
                for r in chain(build_system.get("requires", []), self.BUILD_REQUIRES)
            }
            new = [r for name, r in mandatory.items() if name not in existing]
            for req in reversed(new):
                requirements.insert_line(0, (req,))
            build_system["requires"] = requirements

        return doc

    def move_tests_require(self, doc: R) -> R:
        """Move ``tests_require`` to a ``testing`` extra as optional dependency
        (this option is deprecated in setuptools (the test command is deprecated).

        It assumes ``move_options_missing_in_pep621`` already run (to populate
        ``project:optional-dependencies``.
        """
        if "tests-require" in doc["options"]:
            msg = "The field 'tests_require' is deprecated and no longer supported. "
            msg += "Dependencies will be converted to optional (`testing` extra). "
            msg += "You can use a tool like `tox` or `nox` to replace this workflow."
            warnings.warn(msg, DeprecationWarning)  # noqa: B028
            reqs: CommentedList[str] = doc["options"].pop("tests-require")
            if "project:optional-dependencies" not in doc:
                doc["project:optional-dependencies"] = IR(testing=reqs)
                return doc

            opt_deps = doc["project:optional-dependencies"]
            if "testing" not in opt_deps:
                opt_deps["testing"] = reqs

            testing: CommentedList[str] = opt_deps["testing"]
            test_deps = {Requirement(r).name: r for r in reqs.as_list()}
            existing_deps = {Requirement(r).name: r for r in testing.as_list()}
            new = [r for name, r in test_deps.items() if name not in existing_deps]
            for req in new:
                testing.insert_line(len(testing), (req,))

        return doc

    def make_include_package_data_explicit(self, doc: R) -> R:
        options = doc["options"]
        if "include-package-data" not in options:
            # This allows setuptools to decide to change the default from False to True,
            # when adopting PEP 621
            options["include-package-data"] = False

        return doc

    def parse_setup_py_command_options(self, doc: R) -> R:
        """``distutils`` commands can accept arguments from ``setup.cfg`` files.
        This function moves these arguments to their own ``distutils``
        tool-specific sub-table
        """
        sections = list(doc.keys())
        commands = _distutils_commands() - SKIP_COMMAND_SECTIONS
        for k in sections:
            if isinstance(k, str) and k in commands:
                section = self._be.apply_best_effort_to_section(doc[k])
                for option in section:
                    if isinstance(option, str):
                        section.rename(option, self.normalise_key(option))
                doc[k] = section
                doc.rename(k, ("distutils", k))
        return doc

    def split_subtables(self, out: R) -> R:
        """``setuptools`` emulate nested sections (e.g.: ``options.extras_require``)
        which can be directly expressed in TOML via sub-tables.
        """
        sections = [
            k
            for k in out.keys()
            if isinstance(k, str) and (k.startswith("options.") or ":" in k)
        ]
        for section in sections:
            new_key = SECTION_SPLITTER.split(section)
            if section.startswith("options."):
                new_key = ["tool", "setuptools", *new_key[1:]]
            out.rename(section, tuple(new_key))
        return out

    def ensure_pep518(self, doc: R) -> R:
        """:pep:`518` specifies that any other tool adding configuration under
        ``pyproject.toml`` should use the ``tool`` table. This means that the only
        top-level keys are ``build-system``, ``project`` and ``tool``.
        """
        allowed = ("build-system", "project", "tool", "metadata", "options")
        allowed_prefixes = ("options.", "project:")
        for k in list(doc.keys()):
            key = k
            rest: Sequence = ()
            if isinstance(k, tuple) and not isinstance(key, HiddenKey):
                key, *rest = k
            if isinstance(key, HiddenKey):
                continue
            if not (key in allowed or any(key.startswith(p) for p in allowed_prefixes)):
                doc.rename(k, ("tool", key, *rest))
        return doc

    def pep621_transform(self, doc: R) -> R:
        """Rules are applied sequentially and therefore can interfere with the following
        ones. Please notice that renaming is applied after value processing.
        """
        transformations = [
            # --- value processing and type changes ---
            self.apply_value_processing,
            # --- transformations mainly focusing on PEP 621 ---
            self.merge_and_rename_urls,
            self.merge_authors_maintainers_and_emails,
            self.merge_and_rename_long_description_and_content_type,
            self.handle_license,
            self.move_and_split_entrypoints,
            self.move_options_missing_in_pep621,
            self.remove_metadata_not_in_pep621,
            # --- General fixes
            self.fix_extras_require,
            self.rename_script_files,
            self.handle_packages_find,
            self.handle_dynamic,
            self.move_setup_requires,
            self.move_tests_require,
            self.make_include_package_data_explicit,
            # --- distutils ---
            self.parse_setup_py_command_options,
            # --- final steps ---
            self.split_subtables,
            self.ensure_pep518,
        ]
        out = self.template(doc.__class__)
        out.update(doc)
        out.setdefault("metadata", IR())
        out.setdefault("options", IR())
        out = reduce(apply, transformations, out)
        out.rename("metadata", "project", ignore_missing=True)
        out.rename("options", ("tool", "setuptools"), ignore_missing=True)
        return out

    def normalise_keys(self, cfg: R) -> R:
        """Normalise keys in ``setup.cfg``, by replacing aliases with cannonic names
        and replacing the snake case with kebab case.

        .. note:: Although setuptools recently deprecated kebab case in ``setup.cfg``
           ``pyproject.toml`` use it as a convention (as established in :pep:`517`,
           :pep:`518` and :pep:`621`) so this normalisation makes more sense for the
           translation.
        """
        # Normalise for the same convention as pyproject
        for i in range(len(cfg.order)):
            section_name = cfg.order[i]
            if not isinstance(section_name, str):
                continue
            if not any(section_name.startswith(s) for s in SETUPTOOLS_SECTIONS):
                continue
            section = cfg[section_name]
            cfg.rename(section_name, kebab_case(section_name))
            if any(section_name.startswith(s) for s in SKIP_CHILD_NORMALISATION):
                continue
            for j in range(len(section.order)):
                option_name = section.order[j]
                if not isinstance(option_name, str):
                    continue
                section.rename(option_name, self.normalise_key(option_name))
        # Normalise aliases
        metadata = cfg.get("metadata")
        if not metadata:
            return cfg
        for alias, cannonic in self.setupcfg_aliases().items():
            if alias in metadata:
                msg = f"{alias!r} is deprecated. Translating to {cannonic!r} instead."
                warnings.warn(msg, DeprecationWarning)  # noqa: B028
                metadata.rename(alias, cannonic)
        return cfg

    def normalise_key(self, key: str) -> str:
        """Normalise a single key for option"""
        return kebab_case(key)


# ---- Helpers ----


class Directive(dict):
    """Represent a setuptools' setup.cfg directive (e.g 'file:', 'attr:')

    In TOML these directives can be represented as dict-like objects, however in the
    conversion algorithm we need to be able to differentiate between them.
    By inheriting from dict, we can use directive classes interchangeably but also check
    using `isinstance(obj, Directive)`.
    """

    def __init__(self, kind: str, args: Any):
        self.kind = kind
        self.args = Any
        super().__init__(((kind, args),))


def directive(*directives: str, orelse=split_comment):
    """:obj:`~functools.partial` form of :func:`split_directive`"""
    directives = directives or ("file", "attr")
    return partial(split_directive, directives=directives, orelse=orelse)


def split_directive(
    value: str, directives: Sequence[str] = ("file", "attr"), orelse=split_comment
):
    candidates = (d for d in directives if value.strip().startswith(f"{d}:"))
    directive = next(candidates, None)
    if directive is None:
        return orelse(value)

    raw_value = value.lstrip()[len(directive) + 1 :].strip()
    if directive == "file":
        return Directive(directive, split_list_comma(raw_value))
    return Directive(directive, split_comment(raw_value))


def value_error(field: str):
    """Simply raise a :exc:`ValueError` when used as a transformation function"""

    def _fn(value):
        raise ValueError(f"Invalid value for {field!r}: {value!r}")

    return _fn


def _distutils_commands() -> Set[str]:
    try:
        from . import iterate_entry_points

        commands = [ep.name for ep in iterate_entry_points("distutils.commands")]
    except Exception:
        commands = []
    return {*commands, *COMMAND_SECTIONS}


def _ensure_where_list(where):
    if isinstance(where, Commented):
        return where.as_commented_list()

    return [where]


def _add_marker(dep: str, marker: str) -> str:
    joiner = " and " if ";" in dep else "; "
    return joiner.join((dep, marker))


def split_deps(value: str) -> CommentedList:
    """Setuptools seem to accept line continuations for markers
    (with comments in the middle), and that is more difficult to process.
    e.g.: https://github.com/jaraco/zipp
    """
    opts = {"force_multiline": True} if ENV_MARKER.search(value) else {}
    internal: CommentedList[str] = split_list_semi(value, **opts)
    lines = list(internal)
    L = len(lines)
    i = j = 0
    remove = []
    while i < L:
        line = lines[i]
        if line.comment_only() or not line.value:
            i += 1
            continue
        while line.value and line.value[-1].strip()[-1] == "\\":
            comments: List[Tuple[int, str]] = []
            for j in range(i + 1, L):
                # Find the non commented / non empty line
                following = lines[j]
                if following.value_or(None):
                    line = _fuse_lines(line, following)
                    lines[i] = line
                    if len(comments) == 1 and not line.has_comment():
                        # If just one comment was found in between,
                        # use it as a inline comment
                        remove.append(comments[0][0])
                        line.comment = comments[0][1]
                    remove.append(j)
                    i = j
                    break
                if following.comment:
                    # Store the comments, they might be used as inline
                    comments.append((j, following.comment))
        i += 1

    for i in reversed(remove):  # backwards otherwise we lose track of the indexes
        lines.pop(i)

    return CommentedList(lines)


def _fuse_lines(line1: Commented[List[str]], line2: Commented[List[str]]):
    """Fuse 2 lines in a CommentedList that accidentally split a single
    value between them
    """
    values1 = line1.value
    values2 = line2.value
    # Requires line1 and line2 to not be empty
    assert isinstance(values1, list) and isinstance(values2, list)
    keep1, keep2 = values1[:-1], values2[1:]
    shared = values1[-1].strip().strip("\\").strip() + " " + values2[0].strip()
    return Commented(keep1 + [shared] + keep2, line2.comment)


class _ConfusingPackagesConfig(UserWarning):
    _MSG = """Confusing configuration `[options.packages.find]`.

    Original configuration sets both:

    - `[options] packages = ...` as a list of named packages
    - `[options.packages.find]`

    The confusion comes from the fact that `[options.packages.find]` should be used
    with `[options] packages = find:` or `[options] packages = find_namespace:`.

    Conversion will ignore `[options.packages.find]`, as it cannot be written in the
    TOML format when `[options] packages = ...` is already given.
    """
    __doc__ = _MSG

    @classmethod
    def emit(cls, msg=_MSG, stacklevel=1):
        warnings.warn(msg, category=cls, stacklevel=stacklevel + 1)



import re
from functools import partial
from typing import TypeVar

from ..transformations import split_comment, split_kv_pairs, split_list, split_scalar
from ..types import HiddenKey, IntermediateRepr, Translator

M = TypeVar("M", bound=IntermediateRepr)

_SECTION_SPLITTER = re.compile(r"\.|:|\\")
_KEY_SEP = "="


def activate(translator: Translator):
    profile = translator["best_effort"]
    plugin = BestEffort()
    profile.intermediate_processors.append(plugin.process_values)
    profile.help_text = plugin.__doc__ or ""


class BestEffort:
    """Guess option value conversion based on the string format"""

    def __init__(
        self,
        key_sep=_KEY_SEP,
        section_splitter=_SECTION_SPLITTER,
    ):
        self.key_sep = key_sep
        self.section_splitter = section_splitter
        self.split_dict = partial(split_kv_pairs, key_sep=key_sep)

    def process_values(self, doc: M) -> M:
        doc_items = list(doc.items())
        for name, section in doc_items:
            doc[name] = self.apply_best_effort_to_section(section)
            # Separate nested sections
            if self.section_splitter.search(name):
                keys = tuple(self.section_splitter.split(name))
                doc.rename(name, keys)
        return doc

    def apply_best_effort_to_section(self, section: M) -> M:
        options = list(section.items())
        # Convert option values:
        for field, value in options:
            self.apply_best_effort(section, field, value)
        return section

    def apply_best_effort(self, container: M, field: str, value: str):
        if isinstance(field, HiddenKey):
            return
        if not isinstance(value, str):
            return
        lines = value.splitlines()
        if len(lines) > 1:
            if self.key_sep in value:
                container[field] = self.split_dict(value)
            else:
                container[field] = split_list(value)
        elif field.endswith("version"):
            container[field] = split_comment(value)
        else:
            container[field] = split_scalar(value)



# based on https://coverage.readthedocs.io/en/stable/config.html
from collections.abc import MutableMapping
from functools import partial, update_wrapper
from typing import TypeVar

from ..transformations import coerce_scalar, split_list
from ..types import Transformation as T
from ..types import Translator

M = TypeVar("M", bound=MutableMapping)


def activate(translator: Translator):
    plugin = Coverage()
    profile = translator[".coveragerc"]

    fn = partial(plugin.process_values, prefix="")
    update_wrapper(fn, plugin.process_values)
    profile.intermediate_processors.append(fn)
    profile.help_text = plugin.__doc__ or ""

    for file in ("setup.cfg", "tox.ini"):
        translator[file].intermediate_processors.append(plugin.process_values)


class Coverage:
    """Convert settings to 'pyproject.toml' equivalent"""

    PREFIX = "coverage:"
    SECTIONS = ("run", "paths", "report", "html", "xml", "json")
    LIST_VALUES = (
        "exclude_lines",
        "concurrency",
        "disable_warnings",
        "debug",
        "include",
        "omit",
        "plugins",
        "source",
        "source_pkgs",
        "partial_branches",
    )

    def process_values(self, doc: M, sections=SECTIONS, prefix=PREFIX) -> M:
        for name in sections:
            candidates = [
                doc.get(f"{prefix}{name}"),
                doc.get("tool", {}).get("coverage", {}).get(name),
                doc.get(("tool", "coverage"), {}).get(name),
                doc.get(("tool", "coverage", name)),
            ]
            for section in candidates:
                if section:
                    self.process_section(section)
        return doc

    def process_section(self, section: M):
        for field in section:
            fn: T = split_list if field in self.LIST_VALUES else coerce_scalar
            section[field] = fn(section[field])



import sys
from collections.abc import Mapping, MutableMapping
from typing import TYPE_CHECKING, Any, Callable, List, Optional, TypeVar, Union

from .intermediate_repr import (
    KV,
    Commented,
    CommentedKV,
    CommentedList,
    CommentKey,
    HiddenKey,
    IntermediateRepr,
    Key,
    WhitespaceKey,
)

if sys.version_info <= (3, 8):  # pragma: no cover
    # TODO: Import directly when `python_requires = >= 3.8`
    if TYPE_CHECKING:
        from typing_extensions import Protocol
    else:
        # Not a real replacement but allows getting rid of the dependency
        from abc import ABC as Protocol
else:  # pragma: no cover
    from typing import Protocol


R = TypeVar("R", bound=IntermediateRepr)
T = TypeVar("T")
M = TypeVar("M", bound=MutableMapping)

Scalar = Union[int, float, bool, str]  # TODO: missing time and datetime
CoerceFn = Callable[[str], T]
Transformation = Union[Callable[[str], Any], Callable[[M], M]]

TextProcessor = Callable[[str], str]
IntermediateProcessor = Callable[[R], R]


IniLoadsFn = Callable[[str, Mapping], IntermediateRepr]
IReprCollapseFn = Callable[[IntermediateRepr], T]
TomlDumpsFn = IReprCollapseFn[str]


class CLIChoice(Protocol):
    """:meta private:"""

    name: str
    help_text: str


class Profile(Protocol):
    name: str
    help_text: str
    pre_processors: List[TextProcessor]
    intermediate_processors: List[IntermediateProcessor]
    post_processors: List[TextProcessor]


class ProfileAugmentation(Protocol):
    active_by_default: bool
    name: str
    help_text: str

    def fn(self, profile: Profile): ...

    def is_active(self, explicitly_active: Optional[bool] = None) -> bool:
        """``explicitly_active`` is a tree-state variable: ``True`` if the user
        explicitly asked for the augmentation, ``False`` if the user explicitly denied
        the augmentation, or ``None`` otherwise.
        """
        ...


class Translator(Protocol):
    def __getitem__(self, profile_name: str) -> Profile:
        """Create and register (and return) a translation :class:`Profile`
        (or return a previously registered one) (see :ref:`core-concepts`).
        """
        ...

    def augment_profiles(
        self,
        fn: "ProfileAugmentationFn",
        active_by_default: bool = False,
        name: str = "",
        help_text: str = "",
    ):
        """Register a profile augmentation function (see :ref:`core-concepts`).
        The keyword ``name`` and ``help_text`` can be used to customise the description
        featured in ``ini2toml``'s CLI, but when these arguments are not given (or empty
        strings), ``name`` is taken from ``fn.__name__`` and ``help_text`` is taken from
        ``fn.__doc__`` (docstring).
        """
        ...


Plugin = Callable[[Translator], None]
ProfileAugmentationFn = Callable[[Profile], None]


__all__ = [
    "CLIChoice",
    "CommentKey",
    "Commented",
    "CommentedKV",
    "CommentedList",
    "HiddenKey",
    "IniLoadsFn",
    "IntermediateProcessor",
    "IntermediateRepr",
    "Key",
    "KV",
    "Plugin",
    "Profile",
    "ProfileAugmentation",
    "ProfileAugmentationFn",
    "TextProcessor",
    "Translator",
    "TomlDumpsFn",
    "WhitespaceKey",
    "Scalar",
    "CoerceFn",
    "Transformation",
]



"""This module serves as a compatibility layer between API-compatible
style preserving TOML editing libraries (e.g. tomlkit and atoml).
It makes it easy to swap between implementations for testing (by means of search and
replace).
"""

from collections import UserList
from collections.abc import Mapping, MutableSequence, Sequence
from functools import singledispatch
from typing import Iterable, Optional, Tuple, TypeVar, Union, cast

from tomlkit import (
    aot,
    array,
    comment,
    document,
    dumps,
    inline_table,
    item,
    loads,
    nl,
    string,
    table,
)
from tomlkit.items import AoT, Array, InlineTable, Item, String, Table
from tomlkit.toml_document import TOMLDocument

from ..errors import InvalidTOMLKey
from ..types import (
    KV,
    Commented,
    CommentedKV,
    CommentedList,
    CommentKey,
    IntermediateRepr,
    WhitespaceKey,
)

__all__ = [
    "dumps",
    "loads",
    "convert",
    "collapse",
]

T = TypeVar("T", bound=Union[TOMLDocument, Table, InlineTable])

MAX_INLINE_TABLE_LEN = 67
INLINE_TABLE_LONG_ELEM = 10
MAX_INLINE_TABLE_LONG_ELEM = 5
LONG = 120


def convert(irepr: IntermediateRepr) -> str:
    text = dumps(collapse(irepr, root=True))
    return text.strip() + "\n"  # ensure terminating newline (POSIX requirement)


@singledispatch
def collapse(obj, root=False):
    # Catch all
    return obj


@collapse.register(str)
def _collapse_string(obj: str) -> String:
    return _string(obj)


@collapse.register(Commented)
def _collapse_commented(obj: Commented, root=False) -> Item:
    return create_item(obj.value_or(None), obj.comment)


@collapse.register(CommentedList)
def _collapse_commented_list(obj: CommentedList, root=False) -> Array:
    out = array()
    out.multiline(False)  # Let's manually control the whitespace
    multiline = len(obj) > 1

    for entry in obj.data:
        values = entry.value_or([])

        if multiline:
            collapsed_values = [collapse(v) for v in values]
            if collapsed_values or entry.comment:
                out.add_line(*collapsed_values, comment=entry.comment)
            else:
                out.add_line(indent="")  # Empty line
        else:
            for value in values:
                cast(list, out).append(collapse(value))
            if entry.has_comment():
                cast(Item, out).comment(entry.comment)

    if multiline:
        out.add_line(indent="")  # New line before closing brackets

    return out


@collapse.register(CommentedKV)
def _collapse_commented_kv(
    obj: CommentedKV, root=False, *, _key: str = ""
) -> Union[Table, InlineTable]:
    len_key = len(_key)
    _as_dict = obj.as_dict()

    comments = list(obj._iter_comments())
    len_comments = sum(len(c) for c in comments) + 4
    # ^-- extra 4 for `  # `

    heuristic_len = len_key + len_comments + len(str(_as_dict)) + 3
    # ^-- extra 3 for ` = `

    num_elem = len(_as_dict)
    multiline = (
        heuristic_len > MAX_INLINE_TABLE_LEN
        or num_elem > MAX_INLINE_TABLE_LONG_ELEM
        or len(comments) > 1
    )
    out: Union[Table, InlineTable] = table() if multiline else inline_table()

    for entry in obj.data:
        values = (v for v in entry.value_or([cast(KV, ())]) if v)
        k: Optional[str] = None  # if the for loop is empty, k = None
        for value in values:
            k, v = value
            out[cast(str, k)] = collapse(v)
        if not entry.has_comment():
            continue
        if not multiline:
            cast(Item, out).comment(entry.comment)
            return out
        if k:
            out[k].comment(entry.comment)
        else:
            out.add(comment(entry.comment))
    return out


@collapse.register(IntermediateRepr)
def _collapse_irepr(obj: IntermediateRepr, root=False):
    # guess a good repr
    if root:
        return _convert_irepr_to_toml(obj, document())

    if any(
        v and isinstance(v, (list, Mapping, UserList)) or isinstance(k, CommentKey)
        for k, v in obj.items()
    ):
        return _convert_irepr_to_toml(obj, table())
    else:
        return _convert_irepr_to_toml(obj, inline_table())


@collapse.register(dict)
def _collapse_dict(obj: dict, root=False) -> Union[Table, InlineTable]:
    if not obj:
        return inline_table()
    out: Union[Table, InlineTable] = (
        table()
        if any(v and isinstance(v, (list, dict)) for v in obj.values())
        or len(repr(obj)) > LONG  # simple heuristic
        else inline_table()
    )
    for key, value in obj.items():
        out.append(key, collapse(value))
    return out


@collapse.register(list)
def _collapse_list(obj: list, root=False) -> Union[AoT, Array]:
    is_aot, max_len, total_len, has_nl, has_nested, num_elem = classify_list(obj)

    # Just some heuristics for which kind of array we are going to use
    not_singleline = (
        has_nl
        or max_len > LONG
        or (max_len > INLINE_TABLE_LONG_ELEM and num_elem > MAX_INLINE_TABLE_LONG_ELEM)
    )

    if is_aot and (not_singleline or has_nested):
        return create_aot(collapse(e) for e in obj)

    out = array()
    cast(MutableSequence, out).extend(collapse(e) for e in obj)
    if not_singleline or total_len > LONG:
        out.multiline(True)
    return out


def _convert_irepr_to_toml(irepr: IntermediateRepr, out: T) -> T:
    if irepr.inline_comment and isinstance(out, Item):
        out.comment(irepr.inline_comment)
    for key, value in irepr.items():
        if isinstance(key, WhitespaceKey):
            out.add(nl())
        elif isinstance(key, CommentKey):
            out.add(comment(value))
        elif isinstance(key, tuple):
            parent_key, *rest = key
            if not isinstance(parent_key, str):
                raise InvalidTOMLKey(key)
            if len(rest) == 1:
                nested_key = rest[0]
                collapsed_value = collapse(value)
                collapsed_str = f"{nested_key} = {dumps(collapsed_value)}"
                simplified_str = collapsed_str.replace("= {}", "").replace("= []", "")
                # Force inline table for the simplest cases
                if (
                    not isinstance(collapsed_value, (Table, AoT))
                    and len(collapsed_str) < LONG
                    and "\n" not in collapsed_str.strip()
                    and "{" not in simplified_str
                    and simplified_str.count("=") <= 1
                    # ^-- avoid nested inline-table, unless it is empty
                ):
                    child = out.setdefault(parent_key, inline_table())
                    child[nested_key] = collapsed_value
                    cmt = list(getattr(value, "_iter_comments", lambda: [])())
                    if cmt:
                        child.comment(" ".join(cmt))
                    continue
            else:
                nested_key = tuple(rest)
            p = out.setdefault(parent_key, {})
            if not isinstance(p, (Table, InlineTable)):
                msg = f"Value for `{parent_key}` expected to be a table, found {p!r}"
                raise ValueError(msg)
            _convert_irepr_to_toml(IntermediateRepr({nested_key: value}), p)
        elif isinstance(key, (int, str)):
            _key = str(key)
            if isinstance(value, CommentedKV):
                # Heuristic for better inline tables
                out[str(key)] = _collapse_commented_kv(value, _key=_key)
            elif isinstance(value, IntermediateRepr):
                p = out.setdefault(_key, {})
                _convert_irepr_to_toml(value, p)
            else:
                out[_key] = collapse(value)
    return out


# --- Helpers ---


def create_item(value, comment):
    obj = _item(value)
    if comment is not None:
        obj.comment(comment)
    return obj


def create_table(m: Mapping) -> Table:
    if isinstance(m, Table):
        return m
    t = table()
    for k, v in m.items():
        t.add(k, v)
    return t


def create_aot(elements: Iterable[Mapping]) -> AoT:
    if isinstance(elements, AoT):
        return elements
    out = aot()
    cast(MutableSequence, out).extend(create_table(t) for t in elements)
    return out


def classify_list(seq: Sequence) -> Tuple[bool, int, int, bool, bool, int]:
    """Expensive method that helps to choose what is the best TOML representation
    for a Python list.

    The return value is composed by 6 values, in order:
    - aot(bool): ``True`` if all elements are dict-like objects
    - max_len(int): Rough (and definitely not precise) estimative of the number of
      chars the TOML representation for the largest element would be.
    - total_len(int): Rough (and definitely not precise) estimate of the total number of
      chars in the TOML representation if it was a single line
    - has_nl(bool): if any TOML representation for the elements has a ``\\n`` char.
    - has_nested(bool): if any element has a nested table or array
    - num_elements(int): number of elements in the list.
    """
    is_aot = True
    has_nl = False
    has_nested = False
    max_len = 0
    total_len = 0
    for elem in seq:
        if not isinstance(elem, Mapping):
            is_aot, elem_repr = False, repr(elem)
        else:
            elem_repr = repr(dict(elem.items()))
            simplified_str = elem_repr[1:-1].replace("{}", "").replace("[]", "")
            # ^-- ignore empty inline tables / arrays
            has_nested = has_nested or any(c in simplified_str for c in "{[")
        len_elem = len(elem_repr)
        total_len += len_elem + 2
        max_len = max(max_len, len_elem)
        has_nl = has_nl or "\n" in elem_repr

    return is_aot, max_len, total_len, has_nl, has_nested, len(seq)


def _item(obj) -> Item:
    if isinstance(obj, str):
        return _string(obj)

    return item(obj)


def _string(obj: str) -> String:
    """Try to guess the best TOML representation for a string"""
    multiline = "\n" in obj
    single_line = "".join(x.strip().rstrip("\\") for x in obj.splitlines())
    literal = '"' in obj or "\\" in single_line

    if multiline and not obj.startswith("\n"):
        # TOML feature: during parsing a starting newline character in a multi-line
        # string will be stripped away.
        # So we are free to add it, since it is better for reading
        obj = "\n" + obj

    try:
        return string(obj, literal=literal, multiline=multiline)
    except ValueError:
        # Literal strings are not always possible
        return string(obj, multiline=multiline)



# https://docs.pytest.org/en/latest/reference/reference.html#configuration-options
# https://docs.pytest.org/en/latest/reference/customize.html#config-file-formats
import logging
from collections.abc import MutableMapping
from functools import partial
from typing import TypeVar, Union

from ..transformations import coerce_scalar, remove_comments, split_comment, split_list
from ..types import Commented, IntermediateRepr, Translator

R = TypeVar("R", bound=IntermediateRepr)

_logger = logging.getLogger(__name__)

_split_spaces = partial(split_list, sep=" ")
_split_lines = partial(split_list, sep="\n")
# ^ most of the list values in pytest use whitespace separators,
#   but markers/filterwarnings are a special case.


def activate(translator: Translator):
    plugin = Pytest()
    for file in ("setup.cfg", "tox.ini", "pytest.ini"):
        translator[file].intermediate_processors.append(plugin.process_values)
    translator["pytest.ini"].help_text = plugin.__doc__ or ""


class Pytest:
    """Convert settings to 'pyproject.toml' ('ini_options' table)"""

    LINE_SEPARATED_LIST_VALUES = (
        "markers",
        "filterwarnings",
    )
    SPACE_SEPARATED_LIST_VALUES = (
        "norecursedirs",
        "python_classes",
        "python_files",
        "python_functions",
        "required_plugins",
        "testpaths",
        "usefixtures",
    )

    DONT_TOUCH = ("minversion",)

    def process_values(self, doc: R) -> R:
        candidates = [
            (("pytest", "ini_options"), "pytest", doc),
            (("tool", "pytest", "ini_options"), "tool:pytest", doc),
            (("tool", "pytest", "ini_options"), ("tool", "pytest"), doc),
            (("pytest", "ini_options"), "pytest", doc.get("tool", {})),
        ]
        for new_key, old_key, parent in candidates:
            section = parent.get(old_key)
            if section:
                self.process_section(section)
                parent.rename(old_key, new_key)
        return doc

    def process_section(self, section: MutableMapping):
        for field in section:
            if field in self.DONT_TOUCH:
                continue
            if field in self.LINE_SEPARATED_LIST_VALUES:
                section[field] = _split_lines(section[field])
            elif field in self.SPACE_SEPARATED_LIST_VALUES:
                section[field] = _split_spaces(section[field])
            elif hasattr(self, f"_process_{field}"):
                section[field] = getattr(self, f"_process_{field}")(section[field])
            else:
                section[field] = coerce_scalar(section[field])

    def _process_addopts(self, content: str) -> Union[Commented[str], str]:
        # pytest-dev/pytest#12228: pytest maintainers recommend addopts as string.
        # However, it cannot handle embedded comments, so we have to strip them.

        if "\n" not in content:
            # It is easy to handle inline comments for a single line.
            return split_comment(content)

        if "#" not in content:
            return content

        msg = (
            "Stripping comments from `tool.pytest.ini_options.addopts`.\n"
            "This field is recommended to be a string, however it cannot "
            "contain embedded comments (ref: pytest-dev/pytest#12228)."
        )
        _logger.warning(msg)

        return remove_comments(content)



"""The purpose of this module is to "collapse" the intermediate representation of a TOML
document into Python builtin data types (mainly a composition of :class:`dict`,
:class:`list`, :class:`int`, :class:`float`, :class:`bool`, :class:`str`).

This is **not a loss-less** process, since comments are not preserved.
"""

from collections.abc import Mapping, MutableMapping
from functools import singledispatch
from typing import Any, TypeVar

from ..errors import InvalidTOMLKey
from ..types import Commented, CommentedKV, CommentedList, HiddenKey, IntermediateRepr

__all__ = [
    "convert",
    "collapse",
]

M = TypeVar("M", bound=MutableMapping)


def convert(irepr: IntermediateRepr) -> dict:
    return collapse(irepr)


@singledispatch
def collapse(obj):
    # Catch all
    return obj


@collapse.register(Commented)
def _collapse_commented(obj: Commented) -> Any:
    return obj.value_or(None)


@collapse.register(CommentedList)
def _collapse_commented_list(obj: CommentedList) -> list:
    return [collapse(v) for v in obj.as_list()]


@collapse.register(CommentedKV)
def _collapse_commented_kv(obj: CommentedKV) -> dict:
    return {k: collapse(v) for k, v in obj.as_dict().items()}


@collapse.register(Mapping)
def _collapse_mapping(obj: Mapping) -> dict:
    return _convert_irepr_to_dict(obj, {})


@collapse.register(list)
def _collapse_list(obj: list) -> list:
    return [collapse(e) for e in obj]


def _convert_irepr_to_dict(irepr: Mapping, out: M) -> M:
    for key, value in irepr.items():
        if isinstance(key, HiddenKey):
            continue
        elif isinstance(key, tuple):
            parent_key, *rest = key
            if not isinstance(parent_key, str):
                raise InvalidTOMLKey(key)
            p = out.setdefault(parent_key, {})
            if not isinstance(p, MutableMapping):
                msg = f"Value for `{parent_key}` expected to be Mapping, found {p!r}"
                raise ValueError(msg)
            nested_key = rest[0] if len(rest) == 1 else tuple(rest)
            _convert_irepr_to_dict({nested_key: value}, p)
        elif isinstance(key, (int, str)):
            out[str(key)] = collapse(value)
    return out
